package main

import (
	"errors"
	"fmt"
	"log"
	"math/rand"
	"sync"
	"testing"
)

// 本文ではvar ErrRandomFailure = ...や
// type ErrRandomFailureでエラーを公開する方法を示したので、
// サンプルでは型や変数を隠蔽したい、それでもエラーを判断できるように
// したい、という場合の手法をのひとつ記しておきます。

// 本サンプルでは非公開のerrRandomFailureという型を宣言しています。
// 非公開ではユーザが参照できないのですが、空のerrRandomFailureエラーを作成する
// 関数だけを公開しておきます。
// さらに、同関数は呼び出されるたびにエラーオブジェクトを作成するのではなく、
// 毎回同じインスタンスを返すことによって効率化をはかっています
var xerrRandomFailure = &errRandomFailure{}

func ErrRandomFailure() error {
	return xerrRandomFailure
}

type errRandomFailure struct {
	msg string
}

func (e *errRandomFailure) Error() string {
	return e.msg
}

func (e *errRandomFailure) Is(target error) bool {
	_, ok := target.(*errRandomFailure)
	return ok
}

func TestCollectErrors(t *testing.T) {
	runGoroutines := func() error {
		const max = 10
		var wg sync.WaitGroup

		// バッファを指定することによりデッドロックを避ける
		errch := make(chan error, max)
		wg.Add(max)
		for i := 0; i < max; i++ {
			i := i
			go func(id int, ch chan error) {
				defer wg.Done()
				if rand.Float32() >= 0.7 {
					// 注：ここはエラー定義と同じパッケージ内で利用している想定なので、
					// 非公開の型を自由に使えます
					ch <- &errRandomFailure{
						msg: fmt.Sprintf(`goroutine %d returning error`, id),
					}
				}
			}(i, errch)
		}

		wg.Wait()

		close(errch)

		var errlist []error
		for err := range errch {
			errlist = append(errlist, err)
		}

		return errors.Join(errlist...)
	}

	if err := runGoroutines(); err != nil {
		if errors.Is(err, ErrRandomFailure()) {
			log.Printf("v is ErrRandomFailure = %s", err)
		}
	}
}
